from __future__ import annotations
from typing import Set, Dict, Union, Tuple, TypeVar, List
from common import Terminal, Nonterminal, Character, State, Eps
from nfa import NFA
from grammars_cfg import CFG

# Typecheck is sufficient for grammar being a grammar of certain type
# with exception of Eps only within the init rule and not use of init elsewhere.


class RegGrammar:
    Rules = Dict[Nonterminal, Set[Union[Terminal, Tuple[Terminal, Nonterminal]]]]
    type_var = TypeVar('type_var')

    def __init__(self, nonterminals: Set[Nonterminal],
                 terminals: Set[Terminal],
                 rules: Rules,
                 init: Nonterminal):
        self.nonterminals = nonterminals
        self.terminals = terminals
        self.rules = rules
        self.init = init
        assert self.check()

    @staticmethod
    def reggrammar_from_cfg(cfg: CFG) -> RegGrammar:
        try:
            if cfg.is_regular():
                return RegGrammar(cfg.nonterminals, cfg.terminals, cfg.rules, cfg.init)
        except: # I don't know
            pass

    # unused formal requirements check, regular grammars specific (rule variants)
    def check(self) -> bool:
        assert len(self.nonterminals) > 0, "empty grammar"

        for nonterminal in self.rules:
            assert nonterminal in self.nonterminals, "unknown nonterminal " + nonterminal.name
            for rule in self.rules[nonterminal]:
                if isinstance(rule, Terminal):
                    assert rule in self.terminals, "unknown terminal " + rule.name
                elif not isinstance(rule, Eps):
                    assert rule[0] in self.terminals, "unknown terminal " + rule[0].name
                    assert rule[1] in self.nonterminals, "unknown nonterminal " + rule[1].name

        assert self.init in self.nonterminals, "init not in nonterminals"

        return True

    def reggrammar_to_nfa(self) -> NFA:

        init = State(self.init.name)
        states: Set[State] = set()
        for nonterminal in self.nonterminals:
            states.add(State(nonterminal.name))

        characters: Set[Character] = set()
        for terminal in self.terminals:
            characters.add(Character(terminal.name))

        final = State("new_final")  # TODO assert that name is unique, as for hell in make_total
        states.add(final)
        transition: Dict[Tuple[State, Character], Set[State]] = dict()

        for nonterminal in self.rules:
            state = State(nonterminal.name)
            for rule in self.rules[nonterminal]:

                # rule A -> a becomes d(A, a) = final
                if isinstance(rule, Terminal):  # TODO and a is not \eps
                    character = Character(rule.name)
                    transition.setdefault((state, character), set()).add(final)

                # rule A -> aB becomes d(A, a) = B
                elif (state, rule[0]) in transition:
                    character = Character(rule[0].name)
                    transition[state, character].add(State(rule[1].name))
                else:
                    character = Character(rule[0].name)
                    transition[state, character] = {State(rule[1].name)}

        # TODO if init -> \eps: nfa.final.add(nfa.init)
        # I need to know how to treat \eps
        nfa = NFA(states, characters, transition, init, {final})
        return nfa